/******************************************************************************\
 * Utilities: Icons.m
 *
 * Useful icon handling functions. Requires "Utilities: Miscellaneous" library.
 *
 * (C) Hipposoft 2009 <ahodgkin@rowing.org.uk>
\******************************************************************************/

#include "Icons.h"
#include "ThumbnailGenerator.h" /* For CANVAS_SIZE only */

/* Static buffer needed by getIconFamilyFromCGImage - better than throwing it
 * on the stack at run-time.
 */

static UInt32 paintBuffer[ CANVAS_SIZE * CANVAS_SIZE ];

/* Local functions */

static OSStatus addImages      ( IconFamilyHandle iconHnd,
                                 CGImageRef       cgImage );

static OSStatus addImage       ( IconFamilyHandle iconHnd,
                                 CGImageRef       cgImage,
                                 CGColorSpaceRef  cgColourSpace,
                                 size_t           width );

static OSStatus addImageOrMask ( IconFamilyHandle iconHnd,
                                 CGImageRef       cgImage,
                                 CGColorSpaceRef  cgColourSpace,
                                 size_t           width,
                                 BOOL             mask );

/******************************************************************************\
 * getFolderIcon()
 *
 * Obtain an CGImageRef containing the standard folder icon. Caller must
 * release this when no longer needed. Returns the largest icon available.
 *
 * In:  N/A
 * Out: CGImageRef for the standard folder icon image, or NULL if something
 *      went wrong.
\******************************************************************************/

CGImageRef getFolderIcon( void )
{
    CGImageSourceRef imageSource = NULL;
    CGImageRef       image       = NULL;

    /* Ask the workspace manager for an icon */

    NSWorkspace * workspace   = [ NSWorkspace sharedWorkspace ];
    NSImage     * folder      =
    [
        workspace iconForFile: @"/Volumes" /* Rarely non-standard! */
    ];

    /* Get a representation of this as NSData, which is toll-free bridged
     * to CFDataRef; create a CGImage using that data.
     */

    NSData * data = NULL;
    if ( folder ) data =
    [
        NSBitmapImageRep TIFFRepresentationOfImageRepsInArray:
        [
            folder representations
        ]
    ];

    if ( data ) imageSource = CGImageSourceCreateWithData
    (
        ( CFDataRef ) data, /* Toll-free bridge */
        NULL
    );

    if ( imageSource )
    {
        /* Find the biggest image (ick - this is probably very slow) */

        CGImageRef  biggestImage = NULL;
        size_t      maxWidth     = 0;
        size_t      maxHeight    = 0;
        size_t      count        = CGImageSourceGetCount( imageSource );

        for ( size_t index = 0; index < count; index ++ )
        {
            image = CGImageSourceCreateImageAtIndex
            (
                imageSource,
                index,
                NULL
            );

            size_t width  = CGImageGetWidth  ( image );
            size_t height = CGImageGetHeight ( image );

            /* If this image is bigger than the previously recorded greatest
             * size for both width and height, then free up any previously
             * recorded biggest image - it's not the biggest anymore! - and
             * store details of the bigger item.
             */

            if ( width > maxWidth && height > maxHeight )
            {

                if ( biggestImage ) CFRelease( biggestImage );

                biggestImage = image;
                maxWidth     = width;
                maxHeight    = height;
            }
        }

        image = biggestImage;
        CFRelease( imageSource );
    }

    return image;
}

/******************************************************************************\
 * createIconFamilyFromCGImage()
 *
 * Given a CGImage, returns an IconFamilyHandle for an icon family containing
 * a 'huge' 512x512 RGBA icon type and several smaller sizes generated by
 * scaling down the input image. Without these smaller sizes, the Finder will
 * run very slowly if displaying a file or folder to which the icon family has
 * been assigned as a custom icon set. Ideally the input CGImage should be
 * that size, but if not, it will be stretched to fit.
 *
 * Requires static data buffer "paintBuffer" (see top of file).
 *
 * Heavily based upon:
 *
 *   http://www.carbondev.com/site/?page=GetIconRefFromCGImage
 *
 * In:  CGImageRef for the image of interest.
 * Out: IconFamilyHandle for the equivalent 512x512 RGBA icon, or NULL if
 *      anything went wrong. Note that the caller is responsible for calling
 *      DisposeHandle() on a non-NULL return value when the icon family is no
 *      longer needed.
\******************************************************************************/

OSStatus createIconFamilyFromCGImage( CGImageRef cgImage, IconFamilyHandle * iconHndRef )
{
    SInt32           theSize;
    IconFamilyHandle iconHnd = NULL;
    OSStatus         err     = memFullErr;

    *iconHndRef = NULL;

    /* Create an icon family handle. An icon family handle is just a fixed
     * (big-endian) header and tagged data.
     */

    theSize = sizeof( OSType ) * 2;
    iconHnd = ( IconFamilyHandle ) NewHandle( theSize );

    require ( iconHnd, bailOut );

    ( *iconHnd )->resourceType = EndianU32_NtoB( kIconFamilyType );
    ( *iconHnd )->resourceSize = EndianU32_NtoB( theSize         );

    /* Add the images and return the handle of the populated icon family */

    err = addImages( iconHnd, cgImage );
    require ( err == noErr, bailOut );

    *iconHndRef = iconHnd;
    return err;

bailOut:

    if ( iconHnd != NULL ) DisposeHandle( ( Handle ) iconHnd );
    return err;
}

/******************************************************************************\
 * hasCustomIcon()
 *
 * Does the given file or folder appear to have a custom icon defined? Code is
 * based on the implementation for "saveCustomIcon" so see that for further
 * background information.
 *
 * In:  Full POSIX path of file or folder to update;
 *
 * Out: YES if there's a custom icon resource, NO for any other condition
 *      (including internal errors).
\******************************************************************************/

BOOL hasCustomIcon( NSString * fullPosixPath )
{
    BOOL         result = NO;
    OSStatus     err;
    Boolean      dir;
    FSRef        ref;
    HFSUniStr255 fork = { 0, };
    SInt16       refNum = kResFileNotOpened;

    /* Get an FSRef to refer to the file of interest */

    err = FSPathMakeRef
    (
        ( UInt8 * ) [ fullPosixPath fileSystemRepresentation ], &ref, &dir
    );

    require( err == noErr, bailOut );

    /* If it's a directory, change to looking for the resource file inside */

    if ( dir )
    {
        Boolean    ignored;
        NSString * iconPath =
        [
            fullPosixPath stringByAppendingPathComponent: @"Icon\r"
        ];

        err = FSPathMakeRef
        (
            ( const UInt8 * ) [ iconPath UTF8String ],
            &ref,
            &ignored
        );

        require( err == noErr, bailOut );
    }

    /* Open the resource fork; if this fails, assume no resource fork and
     * bail out with the default result of no custom icon.
     */

    FSGetResourceForkName( &fork );

    err = FSOpenResourceFile
    (
        &ref,
        fork.length,
        fork.unicode,
        fsRdPerm,
        &refNum
    );

    require( err == noErr, bailOut );

    /* Provided that the resources file was opened, check for the 'icns'
     * resource explicitly and if found, note that there was a custom icon
     * present. Tidy up and close the resources file.
     */

    if ( refNum != kResFileNotOpened )
    {
        UseResFile( refNum );

        Handle h = Get1Resource( 'icns', kCustomIconResource );
        if ( h != NULL ) result = YES;

        CloseResFile( refNum ); /* This disposes of handle 'h' in passing */
    }

bailOut:

    return result;
}

/******************************************************************************\
 * saveCustomIcon()
 *
 * Saves a custom icon for the given file. Note that custom disk (volume) icons
 * are not supported, but custom icons for regular files and folders should
 * work fine.
 *
 * The function is a hybrid based on adaptation and a combination of code at:
 *
 *   http://developer.apple.com/samplecode/SetCustomIcon/
 *   http://www.cocoabuilder.com/archive/message/cocoa/2004/2/3/95889
 *
 * In:  Full POSIX path of file or folder to update;
 *
 *      IconFamilyHandle for the icon to set (see "getIconFamilyFromCGImage"
 *      for one of many different ways to obtain such a thing).
 *
 * Out: Error indication - noErr if OK, else failed.
\******************************************************************************/

OSStatus saveCustomIcon( NSString * fullPosixPath, IconFamilyHandle icnsH )
{
    OSStatus      err;
    FSCatalogInfo info;
    FSRef         par;
    FSRef         ref;
    Boolean       dir = false;

    err = FSPathMakeRef
    (
        ( UInt8 * ) [ fullPosixPath fileSystemRepresentation ], &par, &dir
    );

    require( err == noErr, bailOut );

    HFSUniStr255 fork   = { 0, };
    SInt16       refNum = kResFileNotOpened;

    /* Create a resource fork to hold the icon; folders do this using a file
     * called "ICON\r".
     */

    FSGetResourceForkName( &fork );

    if ( dir )
    {
        UniChar iconName[ 5 ] = { 'I', 'c', 'o', 'n', '\r' };

        /* Make the file invisible. The next bit comes from Icon Family -
         * it reckons the type and creator should be set.
         */

        BlockZero( &info, sizeof( info ) );

        struct FileInfo * finderInfo = ( ( FileInfo * ) ( &info.finderInfo ) );

        finderInfo->finderFlags = kIsInvisible;
        finderInfo->fileType    = 'icon';
        finderInfo->fileCreator = 'MACS';

        err = FSCreateResourceFile
        (
            &par,
            5,
            iconName,
            kFSCatInfoFinderInfo,
            &info,
            fork.length,
            fork.unicode,
            &ref,
            NULL
        );

        if ( err == dupFNErr )
        {
            /* Resource already exists - trouble is, 'ref' has not been
             * filled in yet; must do that now.
             */

            Boolean    ignored;
            NSString * iconPath =
            [
                fullPosixPath stringByAppendingPathComponent: @"Icon\r"
            ];

            err = FSPathMakeRef
            (
                ( const UInt8 * ) [ iconPath UTF8String ],
                &ref,
                &ignored
            );
        }
    }
    else
    {
        BlockMoveData( &par, &ref, sizeof( FSRef ) );

        err = FSCreateResourceFork
        (
            &ref,
            fork.length,
            fork.unicode,
            0
        );
    }

    /* "dupFNErr" is allowed - means resource fork already exists */

    require( err == noErr || err == dupFNErr, bailOut );

    /* Open the resource file ready for writing */

    err = FSOpenResourceFile
    (
        &ref,
        fork.length,
        fork.unicode,
        fsRdWrPerm,
        &refNum
    );

    require( err == noErr, bailOut );

    if ( refNum == kResFileNotOpened )
    {
        err = readErr;
        goto bailOut;
    }

    /* Then, with this resource file... */

    UseResFile( refNum );

    /* If it already has a custom icon, remove it first */

    Handle h = Get1Resource( 'icns', kCustomIconResource );

    if ( h != NULL )
    {
        RemoveResource( h );
        DisposeHandle( h );
    }

    /* Create a new resource ("icns", -16455). Add the resource,
     * write the resource and detach it (we can't alter it but we
     * can still refer to it later).
     */

    AddResource( ( Handle ) icnsH, 'icns', kCustomIconResource, "\p" );
    err = ResError();
    require( err == noErr, bailOut );

    WriteResource( ( Handle ) icnsH );
    err = ResError();
    require( err == noErr, bailOut );

    DetachResource( ( Handle ) icnsH );
    err = ResError();
    require( err == noErr, bailOut );

    /* All done - close the resource fork */

    CloseResFile( refNum );

    /* The following code adds volume icon support but uses the deprecated API
     * so is compiled out for now. Maybe I'll get around to writing an updated
     * equivalent one day (...unlikely though!).
     */

#if 0

    if ( targetSpec->parID == fsRtParID )                                                   //  As of Mac OS X 10.4, the disk icon requires it be stored in ".VolumeIcon.icns"
    {
        FSSpec      volumeIconSpec;
        long        ioCount;
        refNum  = 0;
        err = FSMakeFSSpec( targetSpec->vRefNum, fsRtDirID, "\p.VolumeIcon.icns", &volumeIconSpec );    //  "." files are always invisible
        if ( err == noErr ) (void) FSpDelete( &volumeIconSpec );
        err = FSpCreate( &volumeIconSpec, 0, 0, smSystemScript );           if ( err != noErr ) goto BailOnTenFour;
        err = FSpOpenDF( &volumeIconSpec, fsWrPerm, &refNum );              if ( err != noErr ) goto BailOnTenFour;
        ioCount = GetHandleSize( (Handle) icnsH );
        err = FSWrite( refNum, &ioCount, *icnsH );                          if ( err != noErr ) goto BailOnTenFour; //  Write the icns file
BailOnTenFour:
        if ( refNum != 0 )  FSClose( refNum );
    }

#endif

    /* Set the 'has custom icon' attribute */

    err = FSGetCatalogInfo
    (
        &par,
        kFSCatInfoFinderInfo,
        &info,
        NULL,
        NULL,
        NULL
    );

    require( err == noErr, bailOut );

    ( ( FileInfo * ) ( &info.finderInfo ) )->finderFlags = kHasCustomIcon;

    err = FSSetCatalogInfo( &par, kFSCatInfoFinderInfo, &info );
    require( err == noErr, bailOut );

    /* Create an alias to the target file/folder and use it to send the
     * Finder a kAESync AppleEvent so that it updates its display(s) of
     * the target item immediately, if necessary.
     */

    AliasHandle aliasH;
    err = FSNewAliasMinimal( &par, &aliasH );
    require_noerr( err, bailOut );
    err = sendFinderAppleEvent( aliasH, kAESync );
    DisposeHandle( (Handle) aliasH );

bailOut:

    return err;
}

/******************************************************************************\
 * addImages()
 *
 * Internal - to the icon family identified by the given handle, add the
 * given image in various different sizes.
 *
 * In:  Icon family handle;
 *
 *      Reference to a Core Graphics image which should be 512x512 pixels in
 *      size (other image sizes will work, but will be inefficient).
 *
 * Out: OSStatus indication of error ('noErr' if all goes well).
\******************************************************************************/

static OSStatus addImages( IconFamilyHandle iconHnd,
                           CGImageRef       cgImage )
{
    CGColorSpaceRef cgColourSpace = CGColorSpaceCreateDeviceRGB();
    if ( cgColourSpace == NULL ) return memFullErr;

    /* Technically we could just generate a 512x512 icon image but if assigned
     * to a file or folder this makes the Finder run extremely slowly. It seems
     * that it really hates not having smaller icon variants present in the
     * resource fork. Consequently we do all that work up-front, even though
     * this means the icon family generation takes longer. Run-time speed in
     * the Finder thereafter is far more important.
     *
     * We need 512x512 and 256x256 RGBA icons; 128x128, 32x32 and 16x16 icons
     * with 8-bit alpha and RGB data in separate channels; and 32x32 and 16x16
     * icons with 1-bit alpha and 8-bit colour. for a full complement. In
     * practice, a minimum set of 512x512, 256x256 and 128x128 shows the same
     * performance as a full icon set; without 128x128 the Finder will become
     * very slow; with it, the Finder is faster; but for decent speed, both the
     * 512x512 and 256x256 size icons need to be omitted. This seems to be a
     * problem in the Finder itself.
     *
     * Note that display speed issues apply to icon mode only - cover flow,
     * column, and search result view modes show no speed issues at all.
     */

    OSStatus err = noErr;

    if ( err == noErr ) err = addImage( iconHnd, cgImage, cgColourSpace, 512 );
    if ( err == noErr ) err = addImage( iconHnd, cgImage, cgColourSpace, 256 );
    if ( err == noErr ) err = addImage( iconHnd, cgImage, cgColourSpace, 128 );
    // if ( err == noErr ) err = addImage( iconHnd, cgImage, cgColourSpace,  32 );
    // if ( err == noErr ) err = addImage( iconHnd, cgImage, cgColourSpace,  16 );

    CFRelease( cgColourSpace );
    return err;
}

/******************************************************************************\
 * addImage()
 *
 * Internal - to the icon family identified by the given handle, add the
 * given image by painting using the given colour space at the given width.
 * The icon height is set to the same value. Widths of 512, 256, 128, 32 and
 * 16 are supported. Other widths will provoke undefined behaviour.
 *
 * In:  Icon family handle;
 *
 *      Reference to a Core Graphics image which should be 512x512 pixels in
 *      size (other image sizes will work, but will be inefficient);
 *
 *      Reference to a colour space in which to work (usually created via a
 *      call to 'CGColorSpaceCreateDeviceRGB');
 *
 *      Width of icon to add in pixels. Height is set to match. Must be one
 *      value out of 512, 256, 128, 32 or 16.
 *
 * Out: OSStatus indication of error ('noErr' if all goes well).
\******************************************************************************/

static OSStatus addImage( IconFamilyHandle iconHnd,
                          CGImageRef       cgImage,
                          CGColorSpaceRef  cgColourSpace,
                          size_t           width )
{
    OSStatus err = addImageOrMask( iconHnd, cgImage, cgColourSpace, width, NO );

    /* Note that this is a bit inefficient for images that need separate masks
     * since the source Core Graphics image will be scaled to the same size
     * twice, once while painting the image and once while painting its mask.
     */

    if ( err == noErr && width < 256 )
    {
        addImageOrMask( iconHnd, cgImage, cgColourSpace, width, YES );
    }

    return err;
}

/******************************************************************************\
 * addImageOrMask()
 *
 * Internal - as "addImage", but the caller has control of whether an image or
 * image mask gets added to the icon family for widths of 128, 32 or 16. Widths
 * of 256 or 512 use a merged RGBA mask and image data format so both are
 * always added at the same time.
 *
 * Uses the "paintBuffer" static data buffer (see top of file); if the width
 * requested exceeds CANVAS_SIZE, it will be limited to this value to avoid
 * buffer overrun.
 *
 * In:  Icon family handle;
 *
 *      Reference to a Core Graphics image which should be 512x512 pixels in
 *      size (other image sizes will work, but will be inefficient);
 *
 *      Reference to a colour space in which to work (usually created via a
 *      call to 'CGColorSpaceCreateDeviceRGB');
 *
 *      Width of icon to add in pixels. Height is set to match. Must be one
 *      value out of 512, 256, 128, 32 or 16, else behaviour is undefined;
 *
 *      YES to add the image mask (only for widths 128, 32 or 16), else NO.
 *      Using YES for other image widths will provoke undefined behaviour.
 *
 * Out: OSStatus indication of error ('noErr' if all goes well).
\******************************************************************************/

static OSStatus addImageOrMask( IconFamilyHandle iconHnd,
                                CGImageRef       cgImage,
                                CGColorSpaceRef  cgColourSpace,
                                size_t           width,
                                BOOL             mask )
{
    size_t       pixelSize, dataSize;
    CGBitmapInfo info;
    OSType       type;

    /* Figure out the various types and sizes needed */

    if ( mask == YES )
    {
        /* Width of 128, 32 or 16, else undefined results */

        pixelSize = 1;
        info      = kCGImageAlphaOnly;

        switch( width )
        {
            default: type = kThumbnail8BitMask; break; /* 128 or unknown */
            case 32: type = kLarge8BitMask;     break;
            case 16: type = kSmall8BitMask;     break;
        }
    }
    else
    {
        /* Width of 512, 256, 128, 32 or 16, else undefined results */

        pixelSize = 4;

        if ( width >= 256 ) info = kCGImageAlphaPremultipliedFirst;
        else                info = kCGImageAlphaNoneSkipFirst;

        switch( width )
        {
            default:  type = kIconServices512PixelDataARGB; break; /* 512 or unknown */
            case 256: type = kIconServices256PixelDataARGB; break;
            case 128: type = kThumbnail32BitData;           break;
            case 32:  type = kLarge32BitData;               break;
            case 16:  type = kSmall32BitData;               break;
        }
    }

    if ( width > CANVAS_SIZE ) width = CANVAS_SIZE;
    dataSize = width * width * pixelSize;

    /* Clear the paint buffer and create a context within it */

    memset( paintBuffer, '\0', dataSize );

    CGContextRef cgContext = CGBitmapContextCreate
    (
        paintBuffer,
        width,
        width,
        8,
        width * pixelSize,
        ( mask == YES ) ? NULL : cgColourSpace,
        info
    );

    if ( cgContext == NULL ) return memFullErr;

    /* Paint the image */

    CGContextDrawImage( cgContext, CGRectMake( 0, 0, width, width ), cgImage );
    CFRelease( cgContext );

    /* Add it to the icon collection */

    Handle   tmpHnd;
    OSStatus err;

    err = PtrToHand( paintBuffer, &tmpHnd, dataSize );
    if ( err != noErr ) return err;

    err = SetIconFamilyData( iconHnd, type, tmpHnd );
    DisposeHandle( tmpHnd );
    return err;
}
