/******************************************************************************\
 * Utilities: ThumbnailGenerator.m
 *
 * Takes folders and creates a thumbnail icon that gives an idea of any images
 * contained inside that folder. Requires "Utilities: Miscellaneous" library.
 *
 * (C) Hipposoft 2009 <ahodgkin@rowing.org.uk>
\******************************************************************************/

#include "ThumbnailGenerator.h"

/* Declare local functions */

static BOOL paintImage ( CFStringRef  fullPosixPath,
                         CGRect       rect,
                         CGContextRef context );

/* Pre-computed locations inside a CANVAS_SIZE square canvas for cropped
 * thumbnail icons for when there are between 1 and 4 icons available. See
 * "ThumbnailGenerator.h" for CANVAS_SIZE and other relevant definitions.
 *
 * Ordering is important to make sure that when one icon overlaps another,
 * it will overlap with its drop shadow part - this looks much better than
 * an overlap of a white-over-white border region without shadow. Shadows
 * are offset to the right and downwards, so start plotting in the bottom
 * right corner and work anti-clockwise.
 */

static CGRect locations[ 4 ][ 4 ] =
{
    {
        { .origin.x = 112, .origin.y =  76, .size.width = 288, .size.height = 288 }
    },

    {
        { .origin.x = 240, .origin.y = 100, .size.width = 240, .size.height = 240 },
        { .origin.x =  32, .origin.y = 100, .size.width = 240, .size.height = 240 }
    },

    {
        { .origin.x = 236, .origin.y =   0, .size.width = 276, .size.height = 276 },
        { .origin.x =   0, .origin.y =   0, .size.width = 276, .size.height = 276 },
        { .origin.x = 118, .origin.y = 236, .size.width = 276, .size.height = 276 }
    },

    {
        { .origin.x = 236, .origin.y =   0, .size.width = 276, .size.height = 276 },
        { .origin.x =   0, .origin.y =   0, .size.width = 276, .size.height = 276 },
        { .origin.x = 236, .origin.y = 236, .size.width = 276, .size.height = 276 },
        { .origin.x =   0, .origin.y = 236, .size.width = 276, .size.height = 276 }
    }
};

/******************************************************************************\
 * generateThumbnailForFolder()
 *
 * Generate a thumbnail for the folder identified by the given fully specified
 * POSIX-style pathname at CANVAS_SIZE x CANVAS_SIZE resolution (see the
 * "ThumbnailGenerator.h" header file).
 *
 * In:  Fully specified POSIX-style path of folder of interest;
 *
 *      YES to generate an opaque background image (e.g. to use it as a
 *      thumbnail in a QuickLook generator, where thumbnails are always
 *      opaque) or NO to generate a transparent background image (e.g. to use
 *      as a preview in a QuickLook generator, or as a thumbanil in a Quick
 *      Look generator when the 'icon mode' options flag is clear);
 *
 *      YES to not enumerate contents of subdirectories if the subdirectory
 *      is package-like, else enumerate subdirectories regardless.
 *
 * Out: CGImageRef pointing to thumbnail image or NULL on error. If non-NULL,
 *      caller must CFRelease() the memory when finished.
\******************************************************************************/

CGImageRef generateThumbnailForFolder( NSString * fullPosixPath,
                                       BOOL       opaque,
                                       BOOL       skipPackageLikeEntries )
{
    NSAutoreleasePool * pool = [ [ NSAutoreleasePool alloc] init ];
    NSEnumerator      * enumArray;
    NSString          * currFile;
    NSString          * enumPath   = fullPosixPath;
    CGImageRef          finalImage = NULL;

    /**************************************************************************\
     * Search for images and choose up to four at random
    \**************************************************************************/

    srandomdev(); /* Randomise the random number generator */

    /* Look for image files. The search is exited early if a certain number
     * of images have been found or the loop has been running for too long.
     */

    NSDirectoryEnumerator * dirEnum = [
        [ NSFileManager defaultManager ] enumeratorAtPath: enumPath
    ];

    NSMutableArray * images    = [ NSMutableArray arrayWithCapacity: 0 ];
    clock_t          startTime = clock();

    while ( currFile = [ dirEnum nextObject ] )
    {
        NSDictionary * currAttrs = [ dirEnum fileAttributes ];
        if ( ! currAttrs ) continue;

        NSString * fileType = [ currAttrs objectForKey: NSFileType ];
        NSString * fullPath = [
            enumPath stringByAppendingPathComponent: currFile
        ];

        /* Only interested in regular files or directories, nothing more. Skip
         * directories that appear to have filename extensions because
         * they're probably packaged formats of some kind (e.g. applications).
         */

        if ( [ fileType isEqualToString: NSFileTypeRegular ] )
        {
            if ( isImageFile( fullPath ) == YES )
            {
                #ifdef MAXIMUM_IMAGE_SIZE
                {
                    NSNumber * size = [ currAttrs objectForKey: NSFileSize ];
                    if ( [ size unsignedLongLongValue ] > MAXIMUM_IMAGE_SIZE ) continue;
                }
                #endif

                [ images addObject: fullPath ];

                /* Once we have enough images, bail */

                if (
                       MAXIMUM_IMAGES_FOUND != 0 &&
                       [ images count ] >= MAXIMUM_IMAGES_FOUND
                   )
                   break;
            }
        }
        else if (
                       skipPackageLikeEntries
                    && [ fileType isEqualToString: NSFileTypeDirectory ]
                    && isLikeAPackage( fullPath )
                )
        {
            [ dirEnum skipDescendents ];
        }

        /* If we've run for too long, bail */

        if (
               MAXIMUM_LOOP_TIME_TICKS != 0 &&
               clock() - startTime >= MAXIMUM_LOOP_TIME_TICKS
           )
           break;
    }

    /* If there are no images, exit; the standard folder icon will be used */

    require( [ images count ] > 0, nothingToDo );

    /* Otherwise, choose up to four images at random */

    NSMutableArray * chosenImages = [ NSMutableArray arrayWithCapacity: 0 ];

    while ( [ images count ] > 0 && [ chosenImages count ] < 4 )
    {
        /* Using random() in this way is fine - unlike rand(), all bits are
         * considered sufficiently random in the generated integer.
         */

        NSUInteger randomIndex = random() % [ images count ];

        [ chosenImages addObject: [ images objectAtIndex: randomIndex ] ];
        [ images removeObjectAtIndex: randomIndex ];
    }

    /**************************************************************************\
     * Create layers representing thumbnails of the images
    \**************************************************************************/

    /* Get a graphics context for painting things. This is constructed as a
     * bespoke bitmap context rather than using the one we could obtain from
     * Quick Look because it opens up more possibilities later (there is
     * direct control over alpha channel provision, though transparency is
     * extremely problematic under Quick Look - most of the time, you don't
     * get it, regardless of context).
     */

    CGSize          pixelSize  = CGSizeMake(       CANVAS_SIZE, CANVAS_SIZE );
    CGRect          pixelRect  = CGRectMake( 0, 0, CANVAS_SIZE, CANVAS_SIZE );
    CGContextRef    context    = NULL;
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();

    if ( colorSpace )
    {
        context = CGBitmapContextCreate
        (
            NULL,            /* OS X 10.3 or later => CG allocates for us */
            CANVAS_SIZE,
            CANVAS_SIZE,
            8,               /* Bits per component */
            CANVAS_SIZE * 4, /* Bytes per row      */
            colorSpace,
            kCGImageAlphaPremultipliedFirst
        );

        CFRelease( colorSpace );
    }

    require( context, haveNoContext );

    CGContextSetShouldAntialias      ( context, true                 );
    CGContextSetInterpolationQuality ( context, kCGInterpolationHigh );

    /* Create layers into which thumbnails will be drawn. These will be
     * scaled down again when drawn into the final icon and will use anti-
     * aliasing then, but we need to keep high quality settings on the
     * layers to make sure that the rotated edges look good.
     */

    CFMutableArrayRef layers = CFArrayCreateMutable
    (
        kCFAllocatorDefault,
        4,
        NULL
    );

    enumArray = [ chosenImages objectEnumerator ];
    while ( currFile = [ enumArray nextObject ] )
    {
        CGLayerRef   layer    = CGLayerCreateWithContext( context, pixelSize, NULL );
        CGContextRef layerCtx = CGLayerGetContext( layer );

        CGContextSetShouldAntialias      ( layerCtx, true                 );
        CGContextSetInterpolationQuality ( layerCtx, kCGInterpolationHigh );

        /* Paint in a transparency layer using a shadow offset to the bottom
         * and right by BLUR_OFFSET with a radius of BLUR_RADIUS.
         */

        CGContextBeginTransparencyLayer( layerCtx, NULL );

        CGContextTranslateCTM ( layerCtx, CANVAS_SIZE / 2, CANVAS_SIZE / 2      );
        CGContextRotateCTM    ( layerCtx, ( ( random() % 300 ) - 150 ) / 2000.0 );

        CGContextSaveGState( layerCtx );
        CGContextSetShadow
        (
            layerCtx,
            CGSizeMake( BLUR_OFFSET, -BLUR_OFFSET ),
            BLUR_RADIUS
        );

        CGContextSetRGBFillColor( layerCtx, 1, 1, 1, 1.0 );
        CGContextFillRect
        (
            layerCtx,
            CGRectMake
            (
                -BORDER_SIZE / 2,
                -BORDER_SIZE / 2,
                BORDER_SIZE,
                BORDER_SIZE
            )
        );

        /* Don't want another shadow under the inset image so restore the
         * pre-shadow context before calling the painting routine..
         */

        CGContextRestoreGState( layerCtx );

        BOOL success = paintImage
        (
            ( CFStringRef ) currFile, /* Toll-free bridge */
            CGRectMake
            (
                -THUMB_SIZE / 2,
                -THUMB_SIZE / 2,
                THUMB_SIZE,
                THUMB_SIZE
            ),
            layerCtx
        );

        CGContextEndTransparencyLayer( layerCtx );

        if ( success ) CFArrayAppendValue( layers, layer );
        else CFRelease( layer );
    }

    /**************************************************************************\
     * Construct the final thumbnail
    \**************************************************************************/

    /* Iterate through any thumbnails, painting them into the right location
     * on the final canvas, freeing layers as we go.
     */

    CFIndex layerCount = CFArrayGetCount( layers );

    if ( layerCount > 0 )
    {
        CGContextBeginTransparencyLayer( context, NULL );

        /* Quick Look enforces an opaque thumbnail. If we use a clear
         * background then it'll (bizarrely) work fine if looking at the
         * index sheet for a collection of folders selected at once for
         * Quick Look, but single folders and in-Finder icons are all
         * opaque; some with the 'page curl' effect, some without.
         *
         * Callers can anticipate this and ask for an opaque thumbnail
         * if Quick Look is being invoked in 'for Finder-style file icon'
         * mode.
         */

        if ( opaque )
        {
            if ( layerCount < 3 ) CGContextSetRGBFillColor( context, 1.00, 1.00, 1.00, 1.0 );
            else                  CGContextSetRGBFillColor( context, 0.62, 0.77, 0.85, 1.0 );

            CGContextFillRect( context, pixelRect );
        }
        else
        {
            CGContextClearRect( context, pixelRect );
        }

        /* For one or two thumbnails only, plot the standard folder icon
         * underneath.
         */

        if ( layerCount < 3 )
        {
            CGImageRef folder = getFolderIcon();

            if ( folder )
            {
                CGContextDrawImage( context, pixelRect, folder );
                CFRelease( folder );
            }
        }

        /* Now plot the thumbnails themselves */

        for ( CFIndex index = 0; index < layerCount; index ++ )
        {
            CGLayerRef layer = ( CGLayerRef ) CFArrayGetValueAtIndex( layers, index );

            CGContextDrawLayerInRect
            (
                context,
                locations[ layerCount - 1 ][ index ],
                layer
            );

            CFRelease( layer );
        }

        CGContextEndTransparencyLayer( context );

        /* Flush out any pending graphics operations and set the thumbnail
         * response for this generator, or set a custom icon for the folder.
         */

        CGContextFlush( context );
        finalImage = CGBitmapContextCreateImage( context );
    }

    /**************************************************************************\
     * Tidy up
    \**************************************************************************/

    CFRelease( layers  );
    CFRelease( context );

nothingToDo:
haveNoContext:

    [ pool release ];

    return finalImage;
}

/******************************************************************************\
 * paintImage()
 *
 * Paint an image found at the given fully specified POSIX-style file path
 * into the given rectangle under the given context. The image is cropped to
 * a square before being painted; if the target rectangle is not also square
 * the image will be distorted when painted.
 *
 * In:  Full POSIX path of the image to load, crop to 1:1 aspect and paint;
 *
 *      Rectangle into which the cropped image should be painted, distorting
 *      its aspect ratio to fit if necessary;
 *
 *      Graphics context for the painting operation.
 *
 * Out: YES if successful, NO if failed.
\******************************************************************************/

static BOOL paintImage( CFStringRef  fullPosixPath,
                        CGRect       rect,
                        CGContextRef context )
{
    BOOL success = YES;

    /* Turn the POSIX path into a URL, the URL into an image source and the
     * image source into an image object based on index 0 from the source
     * file (i.e. for multi-page TIFFs, Icon files etc., take the first of
     * however many sub-images are contained within).
     */

    CGImageSourceRef imageSource = NULL;
    CGImageRef       image       = NULL;
    CFURLRef         url         = CFURLCreateWithFileSystemPath
    (
        kCFAllocatorDefault,
        fullPosixPath,
        kCFURLPOSIXPathStyle,
        false
    );

    if ( url )
    {
        imageSource = CGImageSourceCreateWithURL( url, NULL );
    }

    if ( imageSource )
    {
        image = CGImageSourceCreateImageAtIndex( imageSource, 0, NULL );
    }

    /* Create a sub-image based on a square crop of the original. The idea
     * is to crop the source before scaling it, because potentially scaling
     * is a much slower operation so the smaller the source image the better.
     *
     * If the image is wider than tall (landscape) we take the center crop.
     * If the image is taller than wide (portrait) we take the top part of
     * the image - on average this works well for pictures of people.
     */

    if ( image )
    {
        CGImageRef newImage;

        size_t width  = CGImageGetWidth( image );
        size_t height = CGImageGetHeight( image );

        if ( width > height )
        {
            newImage = CGImageCreateWithImageInRect
            (
                image,
                CGRectMake( ( width - height ) / 2, 0, height, height )
            );

            CFRelease( image );
            image = newImage;
        }
        else if ( height > width )
        {
            newImage = CGImageCreateWithImageInRect
            (
                image,
                CGRectMake( 0, 0, width, width ) /* Image *top* left is (0,0) */
            );

            CFRelease( image );
            image = newImage;
        }
    }

    /* Check 'image' again in case cropping was attempted but failed */

    if ( image ) CGContextDrawImage( context, rect, image );

    /* Make sure everything is released */

    if ( image       ) CFRelease( image       ); else success = NO;
    if ( imageSource ) CFRelease( imageSource ); else success = NO;
    if ( url         ) CFRelease( url         ); else success = NO;

    return success;
}
