/******************************************************************************\
 * addfoldericon - command line program to assign an icon to a folder
 * which gives an indication of any images held inside that folder, or
 * any of its sub-folders.
 *
 * (C) Hipposoft 2009 <ahodgkin@rowing.org.uk>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
\******************************************************************************/

#include <stdio.h>
#include <sysexits.h>

/* Local library code */

#include "Icons.h"
#include "ThumbnailGenerator.h"

/* Some useful strings */

#define PROGRAM_STRING "addfoldericon"
#define VERSION_STRING "1.20"
#define AUTHOR_STRING  "Andrew Hodgkinson"

/* Avoid subdirectories which are package-like (e.g. applications)? */

#define SKIP_PACKAGES YES

/* Local functions */

static void printHelp    ( void );
static void printVersion ( void );

/******************************************************************************\
 * First thing's first...
\******************************************************************************/

int main( int argc, const char * argv[] )
{
    /* See if a correct number of arguments was submitted - if so, read
     * the argument data, else complain.
     */

    const char * dst;

    if ( argc < 2 )
    {
        printVersion();
        printHelp();
        return EX_USAGE;
    }
    else
    {
        dst = argv[ 1 ];
    }

    /* Get an instance of NSString from the UTF-8 encoded C string */

    NSAutoreleasePool * pool          = [ [ NSAutoreleasePool alloc ] init ];
    NSString          * fullPosixPath =
    [
        NSString stringWithCString: dst
                          encoding: NSUTF8StringEncoding
    ];

    /* Generate the thumbnail */

    CGImageRef finalImage = generateThumbnailForFolder
    (
        fullPosixPath,
        NO,
        SKIP_PACKAGES
    );

    require( finalImage, bail );

    /* Apply the thumbnail to the folder */

    IconFamilyHandle iconHnd = NULL;
    OSStatus         status  = createIconFamilyFromCGImage( finalImage, &iconHnd );

    if ( status == noErr && iconHnd != NULL )
    {
        status = saveCustomIcon( fullPosixPath, iconHnd );
        DisposeHandle( ( Handle ) iconHnd );
    }

    CFRelease( finalImage );

bail:

    [ pool release ];

    if ( status != noErr )
    {
        fprintf
        (
            stderr,
            "%s: Failed with error code %d (&%08X)\n",
            PROGRAM_STRING,
            status,
            status
        );

        perror( "Last known POSIX error" );
    }

    return status;
}

/******************************************************************************\
 * Print version and author to stdout,
\******************************************************************************/

static void printVersion (void)
{
    printf("%s version %s by %s\n", PROGRAM_STRING, VERSION_STRING, AUTHOR_STRING);
}

/******************************************************************************\
 * Print help string to stdout,
\******************************************************************************/

static void printHelp (void)
{
    printf("Usage: %s path_to_folder\n", PROGRAM_STRING);
}
